@@ -1,11 +1,12 @@
 /*
 	audio: audio output interface
 
-	copyright ?-2006 by the mpg123 project - free software under the terms of the LGPL 2.1
+	copyright ?-2008 by the mpg123 project - free software under the terms of the LGPL 2.1
 	see COPYING and AUTHORS files in distribution or http://mpg123.org
 	initially written by Michael Hipp
 */
 
+#include <errno.h>
 #include "mpg123app.h"
 #include "common.h"
 #include "buffer.h"
@@ -14,6 +15,8 @@
 #include <sys/wait.h>
 #endif
 
+#include "debug.h"
+
 static int file_write(struct audio_output_struct* ao, unsigned char *bytes, int count)
 {
 	return (int)write(ao->fn, bytes, count);
@@ -32,6 +35,7 @@ static int builtin_get_formats(struct audio_output_struct *ao)
 		return 0;
 	}
 	else if(param.outmode == DECODE_AU) return MPG123_ENC_SIGNED_16|MPG123_ENC_UNSIGNED_8|MPG123_ENC_ULAW_8;
+	else if(param.outmode == DECODE_WAV) return MPG123_ENC_SIGNED_16|MPG123_ENC_UNSIGNED_8|MPG123_ENC_FLOAT_32|MPG123_ENC_SIGNED_32;
 	else return MPG123_ENC_ANY;
 }
 static int builtin_close(struct audio_output_struct *ao)
@@ -92,49 +96,88 @@ audio_output_t* open_fake_module(void)
 	return ao;
 }
 
-/* Open an audio output module */
-audio_output_t* open_output_module( const char* name )
+/* Open an audio output module, trying modules in list (comma-separated). */
+audio_output_t* open_output_module( const char* names )
 {
 	mpg123_module_t *module = NULL;
 	audio_output_t *ao = NULL;
 	int result = 0;
+	char *curname, *modnames;
 
-	if(param.usebuffer) return NULL;
+	if(param.usebuffer || names==NULL) return NULL;
 
 	/* Use internal code. */
 	if(param.outmode != DECODE_AUDIO) return open_fake_module();
 
-	/* Open the module */
-	module = open_module( "output", name );
-	if (module == NULL) return NULL;
+	modnames = strdup(names);
+	if(modnames == NULL)
+	{
+		error("Error allocating memory for module names.");
+		return NULL;
+	}
+	/* Now loop over the list of possible modules to find one that works. */
+	curname = strtok(modnames, ",");
+	while(curname != NULL)
+	{
+		char* name = curname;
+		curname = strtok(NULL, ",");
+		if(param.verbose > 1) fprintf(stderr, "Trying output module %s.\n", name);
+		/* Open the module, initial check for availability+libraries. */
+		module = open_module( "output", name );
+		if(module == NULL) continue;
+		/* Check if module supports output */
+		if(module->init_output == NULL)
+		{
+			error1("Module '%s' does not support audio output.", name);
+			close_module(module);
+			continue; /* Try next one. */
+		}
+		/* Allocation+initialization of memory for audio output type. */
+		ao = alloc_audio_output();
+		if(ao==NULL)
+		{
+			error("Failed to allocate audio output structure.");
+			close_module(module);
+			break; /* This is fatal. */
+		}
 
-	/* Check module supports output */
-	if (module->init_output == NULL) {
-		error1("Module '%s' does not support audio output.", name);
-		close_module( module );
-		return NULL;
+		/* Call the init function */
+		ao->device = param.output_device;
+		ao->flags  = param.output_flags;
+		/* Should I do funny stuff with stderr file descriptor instead? */
+		if(curname == NULL)
+		{
+			if(param.verbose > 1)
+			fprintf(stderr, "Note: %s is the last output option... showing you any error messages now.\n", name);
+		}
+		else ao->auxflags |= MPG123_OUT_QUIET; /* Probing, so don't spill stderr with errors. */
+		ao->is_open = FALSE;
+		ao->module = module; /* Need that to close module later. */
+		result = module->init_output(ao);
+		if(result == 0)
+		{ /* Try to open the device. I'm only interested in actually working modules. */
+			result = open_output(ao);
+			close_output(ao);
+		}
+		else error2("Module '%s' init failed: %i", name, result);
+
+		if(result!=0)
+		{ /* Try next one... */
+			close_module(module);
+			free(ao);
+			ao = NULL;
+		}
+		else 
+		{ /* All good, leave the loop. */
+			if(param.verbose > 1) fprintf(stderr, "Output module '%s' chosen.\n", name);
+
+			ao->auxflags &= ~MPG123_OUT_QUIET;
+			break;
+		}
 	}
-	
-	/* Allocation memory for audio output type */
-	ao = alloc_audio_output();
-	if (ao==NULL) {
-		error( "Failed to allocate audio output structure." );
-		return NULL;
-	}
-	
-	/* Call the init function */
-	ao->device = param.output_device;
-	ao->flags  = param.output_flags;
-	ao->is_open = FALSE;
-	result = module->init_output(ao);
-	if (result) {
-		error1( "Module's init function failed: %d", result );
-		close_module( module );
-		return NULL;
-	}
-	
-	/* Store the pointer to the module (so we can close it later) */
-	ao->module = module;
+
+	free(modnames);
+	if(ao==NULL) error1("Unable to find a working output module in this list: %s", names);
 
 	return ao;
 }
@@ -171,13 +214,14 @@ audio_output_t* alloc_audio_output()
 	/* Initialise variables */
 	ao->fn = -1;
 	ao->rate = -1;
-	ao->gain = -1;
+	ao->gain = param.gain;
 	ao->userptr = NULL;
 	ao->device = NULL;
 	ao->channels = -1;
 	ao->format = -1;
 	ao->flags = 0;
-	
+	ao->auxflags = 0;
+
 	/*ao->module = NULL;*/
 
 	/* Set the callbacks to NULL */
@@ -204,20 +248,59 @@ static void audio_output_dump(audio_output_t *ao)
 }
 */
 
+struct enc_desc
+{
+	int code; /* MPG123_ENC_SOMETHING */
+	const char *longname; /* signed bla bla */
+	const char *name; /* sXX, short name */
+	const unsigned char nlen; /* significant characters in short name */
+};
+
+static const struct enc_desc encdesc[] =
+{
+	{ MPG123_ENC_SIGNED_16, "signed 16 bit", "s16 ", 3 },
+	{ MPG123_ENC_UNSIGNED_16, "unsigned 16 bit", "u16 ", 3 },
+	{ MPG123_ENC_UNSIGNED_8, "unsigned 8 bit", "u8  ", 2 },
+	{ MPG123_ENC_SIGNED_8, "signed 8 bit", "s8  ", 2 },
+	{ MPG123_ENC_ULAW_8, "mu-law (8 bit)", "ulaw ", 4 },
+	{ MPG123_ENC_ALAW_8, "a-law (8 bit)", "alaw ", 4 },
+	{ MPG123_ENC_FLOAT_32, "float (32 bit)", "f32 ", 3 },
+	{ MPG123_ENC_SIGNED_32, "signed 32 bit", "s32 ", 3 },
+	{ MPG123_ENC_UNSIGNED_32, "unsigned 32 bit", "u32 ", 3 }
+};
+#define KNOWN_ENCS (sizeof(encdesc)/sizeof(struct enc_desc))
+
+void audio_enclist(char** list)
+{
+	size_t length = 0;
+	int i;
+	*list = NULL;
+	for(i=0;i<KNOWN_ENCS;++i) length += encdesc[i].nlen;
+
+	length += KNOWN_ENCS-1; /* spaces between the encodings */
+	*list = malloc(length+1); /* plus zero */
+	if(*list != NULL)
+	{
+		size_t off = 0;
+		(*list)[length] = 0;
+		for(i=0;i<KNOWN_ENCS;++i)
+		{
+			if(i>0) (*list)[off++] = ' ';
+			memcpy(*list+off, encdesc[i].name, encdesc[i].nlen);
+			off += encdesc[i].nlen;
+		}
+	}
+}
 
 /* Safer as function... */
 const char* audio_encoding_name(const int encoding, const int longer)
 {
 	const char *name = longer ? "unknown" : "???";
-	switch(encoding)
-	{
-		case MPG123_ENC_SIGNED_16:   name = longer ? "signed 16 bit"   : "s16 ";  break;
-		case MPG123_ENC_UNSIGNED_16: name = longer ? "unsigned 16 bit" : "u16 ";  break;
-		case MPG123_ENC_UNSIGNED_8:  name = longer ? "unsigned 8 bit"  : "u8  ";   break;
-		case MPG123_ENC_SIGNED_8:    name = longer ? "signed 8 bit"    : "s8  ";   break;
-		case MPG123_ENC_ULAW_8:      name = longer ? "mu-law (8 bit)"  : "ulaw "; break;
-		case MPG123_ENC_ALAW_8:      name = longer ? "a-law (8 bit)"   : "alaw "; break;
-	}
+	int i;
+	for(i=0;i<KNOWN_ENCS;++i)
+	if(encdesc[i].code == encoding)
+	name = longer ? encdesc[i].longname : encdesc[i].name;
+
 	return name;
 }
 
@@ -227,7 +310,7 @@ static void capline(mpg123_handle *mh, long rate)
 	const int  *encs;
 	size_t      num_encs;
 	mpg123_encodings(&encs, &num_encs);
-	fprintf(stderr," %5ld  |", rate);
+	fprintf(stderr," %5ld |", pitch_rate(rate));
 	for(enci=0; enci<num_encs; ++enci)
 	{
 		switch(mpg123_format_support(mh, rate, encs[enci]))
@@ -257,9 +340,13 @@ void print_capabilities(audio_output_t *ao, mpg123_handle *mh)
 	}
 	mpg123_rates(&rates, &num_rates);
 	mpg123_encodings(&encs, &num_encs);
-	fprintf(stderr,"\nAudio driver: %s\nAudio device: %s\nAudio capabilities:\n(matrix of [S]tereo or [M]ono support for sample format and rate in Hz)\n        |", name, dev);
+	fprintf(stderr,"\nAudio driver: %s\nAudio device: %s\nAudio capabilities:\n(matrix of [S]tereo or [M]ono support for sample format and rate in Hz)\n       |", name, dev);
 	for(e=0;e<num_encs;e++) fprintf(stderr," %5s |",audio_encoding_name(encs[e], 0));
-	fprintf(stderr,"\n --------------------------------------------------------\n");
+
+	fprintf(stderr,"\n ------|");
+	for(e=0;e<num_encs;e++) fprintf(stderr,"-------|");
+
+	fprintf(stderr, "\n");
 	for(r=0; r<num_rates; ++r) capline(mh, rates[r]);
 
 	if(param.force_rate) capline(mh, param.force_rate);
@@ -271,20 +358,42 @@ void print_capabilities(audio_output_t *ao, mpg123_handle *mh)
    In case of buffered playback, this works _once_ by querying the buffer for the caps before entering the main loop. */
 void audio_capabilities(audio_output_t *ao, mpg123_handle *mh)
 {
+	int force_fmt = 0;
 	int fmts;
 	size_t ri;
-	long rate;
+	/* Pitching introduces a difference between decoder rate and playback rate. */
+	long rate, decode_rate;
 	int channels;
 	const long *rates;
 	size_t      num_rates, rlimit;
 	debug("audio_capabilities");
 	mpg123_rates(&rates, &num_rates);
 	mpg123_format_none(mh); /* Start with nothing. */
+	if(param.force_encoding != NULL)
+	{
+		int i;
+		if(!param.quiet) fprintf(stderr, "Note: forcing output encoding %s\n", param.force_encoding);
+
+		for(i=0;i<KNOWN_ENCS;++i)
+		if(!strncasecmp(encdesc[i].name, param.force_encoding, encdesc[i].nlen))
+		{
+			force_fmt = encdesc[i].code;
+			break;
+		}
+
+		if(i==KNOWN_ENCS)
+		{
+			error1("Failed to find an encoding to match requested \"%s\"!\n", param.force_encoding);
+			return; /* No capabilities at all... */
+		}
+		else if(param.verbose > 2) fprintf(stderr, "Note: forcing encoding code 0x%x\n", force_fmt);
+	}
 	rlimit = param.force_rate > 0 ? num_rates+1 : num_rates;
 	for(channels=1; channels<=2; channels++)
 	for(ri = 0;ri<rlimit;ri++)
 	{
-		rate = ri < num_rates ? rates[ri] : param.force_rate;
+		decode_rate = ri < num_rates ? rates[ri] : param.force_rate;
+		rate = pitch_rate(decode_rate);
 		if(param.verbose > 2) fprintf(stderr, "Note: checking support for %liHz/%ich.\n", rate, channels);
 #ifndef NOXFERMEM
 		if(param.usebuffer)
@@ -304,15 +413,26 @@ void audio_capabilities(audio_output_t *ao, mpg123_handle *mh)
 			ao->channels = channels;
 			fmts = ao->get_formats(ao);
 		}
-		if(param.verbose > 2) fprintf(stderr, "Note: result 0x%x", fmts);
+		if(param.verbose > 2) fprintf(stderr, "Note: result 0x%x\n", fmts);
+		if(force_fmt)
+		{ /* Filter for forced encoding. */
+			if((fmts & force_fmt) == force_fmt) fmts = force_fmt;
+			else fmts = 0; /* Nothing else! */
+
+			if(param.verbose > 2) fprintf(stderr, "Note: after forcing 0x%x\n", fmts);
+		}
 
 		if(fmts < 0) continue;
-		else mpg123_format(mh, rate, channels, fmts);
+		else mpg123_format(mh, decode_rate, channels, fmts);
 	}
 
 #ifndef NOXFERMEM
 	/* Buffer loop shall start normal operation now. */
-	if(param.usebuffer) xfermem_putcmd(buffermem->fd[XF_WRITER], XF_CMD_WAKEUP);
+	if(param.usebuffer)
+	{
+		xfermem_putcmd(buffermem->fd[XF_WRITER], XF_CMD_WAKEUP);
+		xfermem_getcmd(buffermem->fd[XF_WRITER], TRUE);
+	}
 #endif
 
 	if(param.verbose > 1) print_capabilities(ao, mh);
@@ -338,17 +458,6 @@ int init_output(audio_output_t **ao)
 	init_done = TRUE;
   
 #ifndef NOXFERMEM
-	/*
-	* Only DECODE_AUDIO and DECODE_FILE are sanely handled by the
-	* buffer process. For now, we just ignore the request
-	* to buffer the output. [dk]
-	*/
-	if (param.usebuffer && (param.outmode != DECODE_AUDIO) &&
-	(param.outmode != DECODE_FILE)) {
-	fprintf(stderr, "Sorry, won't buffer output unless writing plain audio.\n");
-	param.usebuffer = 0;
-	} 
-
 	if (param.usebuffer)
 	{
 		unsigned int bufferbytes;
@@ -477,7 +586,17 @@ int flush_output(audio_output_t *ao, unsigned char *bytes, size_t count)
 		else
 #endif
 		if(param.outmode != DECODE_TEST)
-		return ao->write(ao, bytes, (int)count);
+		{
+			int sum = 0;
+			int written;
+			do
+			{ /* Be in a loop for SIGSTOP/CONT */
+				written = ao->write(ao, bytes, (int)count);
+				if(written >= 0){ sum+=written; count -= written; }
+				else error1("Error in writing audio (%s?)!", strerror(errno));
+			}	while(count>0 && written>=0);
+			return sum;
+		}
 	}
 	return (int)count; /* That is for DECODE_TEST */
 }
@@ -503,7 +622,7 @@ int open_output(audio_output_t *ao)
 			ao->is_open = ao->open(ao) < 0 ? FALSE : TRUE;
 			if(!ao->is_open)
 			{
-				error("failed to open audio device");
+				if(!AOQUIET) error("failed to open audio device");
 				return -1;
 			}
 			else return 0;
