@@ -1,7 +1,7 @@
 /*
 	mpg123: main code of the program (not of the decoder...)
 
-	copyright 1995-2008 by the mpg123 project - free software under the terms of the LGPL 2.1
+	copyright 1995-2009 by the mpg123 project - free software under the terms of the LGPL 2.1
 	see COPYING and AUTHORS files in distribution or http://mpg123.org
 	initially written by Michael Hipp
 */
@@ -9,6 +9,7 @@
 #define ME "main"
 #include "mpg123app.h"
 #include "mpg123.h"
+#include "local.h"
 
 #ifdef HAVE_SYS_WAIT_H
 #include <sys/wait.h>
@@ -41,9 +42,7 @@
 #include "playlist.h"
 #include "httpget.h"
 #include "metaprint.h"
-#include "getbits.h"
 #include "httpget.h"
-#include "getcpuflags.h"
 
 #include "debug.h"
 
@@ -106,9 +105,15 @@ struct parameter param = {
 	,0.0 /* pitch */
 	,0 /* ignore_mime */
 	,NULL /* proxyurl */
+	,0 /* keep_open */
+	,0 /* force_utf8 */
+	,INDEX_SIZE
+	,NULL /* force_encoding */
+	,1. /* preload */
+	,-1 /* preframes */
+	,-1 /* gain */
 };
 
-int utf8env = 0;
 mpg123_handle *mh = NULL;
 off_t framenum;
 off_t frames_left;
@@ -350,20 +355,17 @@ topt opts[] = {
 	{0,   "mono",        GLO_INT,  set_frameflag, &frameflag, MPG123_MONO_MIX},
 	{0,   "stereo",      GLO_INT,  set_frameflag, &frameflag, MPG123_FORCE_STEREO},
 	{0,   "reopen",      GLO_INT,  0, &param.force_reopen, 1},
-/*	{'g', "gain",        GLO_ARG | GLO_LONG, 0, &ao.gain,    0}, FIXME */
+	{'g', "gain",        GLO_ARG | GLO_LONG, 0, &param.gain,    0},
 	{'r', "rate",        GLO_ARG | GLO_LONG, 0, &param.force_rate,  0},
 	{0,   "8bit",        GLO_INT,  set_frameflag, &frameflag, MPG123_FORCE_8BIT},
+	{0,   "float",       GLO_INT,  set_frameflag, &frameflag, MPG123_FORCE_FLOAT},
 	{0,   "headphones",  0,                  set_output_h, 0,0},
 	{0,   "speaker",     0,                  set_output_s, 0,0},
 	{0,   "lineout",     0,                  set_output_l, 0,0},
 	{'o', "output",      GLO_ARG | GLO_CHAR, set_output, 0,  0},
 	{0,   "list-modules",0,        list_modules, NULL,  0}, 
 	{'a', "audiodevice", GLO_ARG | GLO_CHAR, 0, &param.output_device,  0},
-#ifdef FLOATOUT
-	{'f', "scale",       GLO_ARG | GLO_DOUBLE, 0, &param.outscale,   0},
-#else
 	{'f', "scale",       GLO_ARG | GLO_LONG, 0, &param.outscale,   0},
-#endif
 	{'n', "frames",      GLO_ARG | GLO_LONG, 0, &param.frame_number,  0},
 	#ifdef HAVE_TERMIOS
 	{'C', "control",     GLO_INT,  0, &param.term_ctrl, TRUE},
@@ -432,6 +434,14 @@ topt opts[] = {
 	{0, "resync-limit", GLO_ARG | GLO_LONG, 0, &param.resync_limit, 0},
 	{0, "pitch", GLO_ARG|GLO_DOUBLE, 0, &param.pitch, 0},
 	{0, "ignore-mime", GLO_INT,  0, &param.ignore_mime, 1 },
+	{0, "keep-open", GLO_INT, 0, &param.keep_open, 1},
+	{0, "utf8", GLO_INT, 0, &param.force_utf8, 1},
+	{0, "fuzzy", GLO_INT,  set_frameflag, &frameflag, MPG123_FUZZY},
+	{0, "index-size", GLO_ARG|GLO_LONG, 0, &param.index_size, 0},
+	{0, "no-seekbuffer", GLO_INT, unset_frameflag, &frameflag, MPG123_SEEKBUFFER},
+	{'e', "encoding", GLO_ARG|GLO_CHAR, 0, &param.force_encoding, 0},
+	{0, "preload", GLO_ARG|GLO_DOUBLE, 0, &param.preload, 0},
+	{0, "preframes", GLO_ARG|GLO_LONG, 0, &param.preframes, 0},
 	{0, 0, 0, 0, 0, 0}
 };
 
@@ -675,20 +685,19 @@ int main(int argc, char *argv[])
 #if !defined(WIN32) && !defined(GENERIC)
 	struct timeval start_time;
 #endif
+
+	/* Extract binary and path, take stuff before/after last / or \ . */
+	if((prgName = strrchr(argv[0], '/')) || (prgName = strrchr(argv[0], '\\')))
 	{
-		char *lang = getenv("LANG");
-		if(lang && strstr(lang, "UTF-8")) utf8env = 1;
-	}
-	{
-		/* Hack the path of the binary... needed for relative module search. */
-		int i;
+		/* There is some explicit path. */
+		prgName[0] = 0; /* End byte for path. */
+		prgName++;
 		binpath = argv[0];
-		for(i=strlen(binpath)-1; i>-1; --i)
-		if(binpath[i] == '/' || binpath[i] == '\\')
-		{
-			binpath[i] = 0;
-			break;
-		}
+	}
+	else
+	{
+		prgName = argv[0]; /* No path separators there. */
+		binpath = NULL; /* No path at all. */
 	}
 
 	/* Need to initialize mpg123 lib here for default parameter values. */
@@ -712,21 +721,18 @@ int main(int argc, char *argv[])
 	mpg123_getpar(mp, MPG123_RVA, &param.rva, NULL);
 	mpg123_getpar(mp, MPG123_DOWNSPEED, &param.halfspeed, NULL);
 	mpg123_getpar(mp, MPG123_UPSPEED, &param.doublespeed, NULL);
-#ifdef FLOATOUT
-	mpg123_getpar(mp, MPG123_OUTSCALE, NULL, &param.outscale);
-#else
 	mpg123_getpar(mp, MPG123_OUTSCALE, &param.outscale, NULL);
-#endif
 	mpg123_getpar(mp, MPG123_FLAGS, &parr, NULL);
+	mpg123_getpar(mp, MPG123_INDEX_SIZE, &param.index_size, NULL);
 	param.flags = (int) parr;
+	param.flags |= MPG123_SEEKBUFFER; /* Default on, for HTTP streams. */
 	mpg123_getpar(mp, MPG123_RESYNC_LIMIT, &param.resync_limit, NULL);
+	mpg123_getpar(mp, MPG123_PREFRAMES, &param.preframes, NULL);
 
 #ifdef OS2
         _wildcard(&argc,&argv);
 #endif
 
-	(prgName = strrchr(argv[0], '/')) ? prgName++ : (prgName = argv[0]);
-
 	while ((result = getlopt(argc, argv, opts)))
 	switch (result) {
 		case GLO_UNKNOWN:
@@ -738,10 +744,12 @@ int main(int argc, char *argv[])
 				prgName, loptarg);
 			usage(1);
 	}
+	/* Do this _after_ parameter parsing. */
+	check_locale(); /* Check/set locale; store if it uses UTF-8. */
 
 	if(param.list_cpu)
 	{
-		char **all_dec = mpg123_decoders();
+		const char **all_dec = mpg123_decoders();
 		printf("Builtin decoders:");
 		while(*all_dec != NULL){ printf(" %s", *all_dec); ++all_dec; }
 		printf("\n");
@@ -750,13 +758,17 @@ int main(int argc, char *argv[])
 	}
 	if(param.test_cpu)
 	{
-		char **all_dec = mpg123_supported_decoders();
+		const char **all_dec = mpg123_supported_decoders();
 		printf("Supported decoders:");
 		while(*all_dec != NULL){ printf(" %s", *all_dec); ++all_dec; }
 		printf("\n");
 		mpg123_delete_pars(mp);
 		return 0;
 	}
+	if(param.gain != -1)
+	{
+	    warning("The parameter -g is deprecated and may be removed in the future.");
+	}
 
 	if (loptind >= argc && !param.listname && !param.remote) usage(1);
 	/* Init audio as early as possible.
@@ -789,7 +801,6 @@ int main(int argc, char *argv[])
 	/* Set the frame parameters from command line options */
 	if(param.quiet) param.flags |= MPG123_QUIET;
 
-	param.flags |= MPG123_SEEKBUFFER; /* For HTTP streams. */
 #ifdef OPT_3DNOW
 	if(dnow != 0) param.cpu = (dnow == SET_3DNOW) ? "3dnow" : "i586";
 #endif
@@ -815,13 +826,10 @@ int main(int argc, char *argv[])
 	    && ++libpar
 	    && MPG123_OK == (result = mpg123_par(mp, MPG123_TIMEOUT, param.timeout, 0))
 #endif
-#ifdef FLOATOUT
-	    && ++libpar
-	    && MPG123_OK == (result = mpg123_par(mp, MPG123_OUTSCALE, 0, param.outscale))
-#else
 	    && ++libpar
 	    && MPG123_OK == (result = mpg123_par(mp, MPG123_OUTSCALE, param.outscale, 0))
-#endif
+	    && ++libpar
+	    && MPG123_OK == (result = mpg123_par(mp, MPG123_PREFRAMES, param.preframes, 0))
 			))
 	{
 		error2("Cannot set library parameter %i: %s", libpar, mpg123_plain_strerror(result));
@@ -829,6 +837,13 @@ int main(int argc, char *argv[])
 	}
 	if (!(param.listentry < 0) && !param.quiet) print_title(stderr); /* do not pollute stdout! */
 
+	{
+		long default_index;
+		mpg123_getpar(mp, MPG123_INDEX_SIZE, &default_index, NULL);
+		if( param.index_size != default_index && (result = mpg123_par(mp, MPG123_INDEX_SIZE, param.index_size, 0.)) != MPG123_OK )
+		error1("Setting of frame index size failed: %s", mpg123_plain_strerror(result));
+	}
+
 	if(param.force_rate && param.down_sample)
 	{
 		error("Down sampling and fixed rate options not allowed together!");
@@ -847,28 +862,7 @@ int main(int argc, char *argv[])
 	/* Now either check caps myself or query buffer for that. */
 	audio_capabilities(ao, mh);
 
-	if(equalfile != NULL)
-	{ /* tst; ThOr: not TRUE or FALSE: allocated or not... */
-		FILE *fe;
-		int i;
-		fe = fopen(equalfile,"r");
-		if(fe) {
-			char line[256];
-			for(i=0;i<32;i++) {
-				float e1,e0; /* %f -> float! */
-				line[0]=0;
-				fgets(line,255,fe);
-				if(line[0]=='#')
-					continue;
-				sscanf(line,"%f %f",&e0,&e1);
-				mpg123_eq(mh, MPG123_LEFT,  i, e0);
-				mpg123_eq(mh, MPG123_RIGHT, i, e1);
-			}
-			fclose(fe);
-		}
-		else
-			fprintf(stderr,"Can't open equalizer file '%s'\n",equalfile);
-	}
+	load_equalizer(mh);
 
 #ifdef HAVE_SETPRIORITY
 	if(param.aggressive) { /* tst */
@@ -950,7 +944,15 @@ int main(int argc, char *argv[])
 			if(param.verbose) fprintf(stderr, "indexing...\r");
 			mpg123_scan(mh);
 		}
+		/*
+			Only trigger a seek if we do not want to start with the first frame.
+			Rationale: Because of libmpg123 sample accuracy, this could cause an unnecessary backwards seek, that even may fail on non-seekable streams.
+			For start frame of 0, we are already at the correct position!
+		*/
+		framenum = 0;
+		if(param.start_frame > 0)
 		framenum = mpg123_seek_frame(mh, param.start_frame, SEEK_SET);
+
 		if(framenum < 0)
 		{
 			error1("Initial seek failed: %s", mpg123_strerror(mh));
@@ -958,7 +960,7 @@ int main(int argc, char *argv[])
 			{
 				fprintf(stderr, "%s", "So, you have trouble getting an output format... this is the matrix of currently possible formats:\n");
 				print_capabilities(ao, mh);
-				fprintf(stderr, "%s", "Somehow the input data and your choices don't allow one of these.\nTried enabling resampling with a forced rate (only 2to1 and 4to1 downsampling is tried automatically)?\nTried allowing mono/stereo?\n");
+				fprintf(stderr, "%s", "Somehow the input data and your choices don't allow one of these.\n");
 			}
 			mpg123_close(mh);
 			continue;
@@ -970,6 +972,12 @@ int main(int argc, char *argv[])
 				&dirname, &filename))
 				fprintf(stderr, "Directory: %s\n", dirname);
 
+#ifdef HAVE_TERMIOS
+		/* Reminder about terminal usage. */
+		if(param.term_ctrl) term_hint();
+#endif
+
+
 			fprintf(stderr, "Playing MPEG stream %lu of %lu: %s ...\n", (unsigned long)pl.pos, (unsigned long)pl.fill, filename);
 			if(htd.icy_name.fill) fprintf(stderr, "ICY-NAME: %s\n", htd.icy_name.p);
 			if(htd.icy_url.fill)  fprintf(stderr, "ICY-URL: %s\n",  htd.icy_url.p);
@@ -1088,8 +1096,8 @@ static void usage(int err)  /* print syntax & exit */
 	fprintf(o,"   -w <filename> write Output as WAV file\n");
 	fprintf(o,"   -k n  skip first n frames [0]        -n n  decode only n frames [all]\n");
 	fprintf(o,"   -c    check range violations         -y    DISABLE resync on errors\n");
-	fprintf(o,"   -b n  output buffer: n Kbytes [0]    -f n  change scalefactor [%g]\n", (double)param.outscale);
-	fprintf(o,"   -r n  set/force samplerate [auto]    -g n  set audio hardware output gain\n");
+	fprintf(o,"   -b n  output buffer: n Kbytes [0]    -f n  change scalefactor [%li]\n", param.outscale);
+	fprintf(o,"   -r n  set/force samplerate [auto]\n");
 	fprintf(o,"   -os,-ol,-oh  output to built-in speaker,line-out connector,headphones\n");
 	#ifdef NAS
 	fprintf(o,"                                        -a d  set NAS server\n");
@@ -1122,6 +1130,7 @@ static void want_usage(char* arg)
 
 static void long_usage(int err)
 {
+	char *enclist;
 	FILE* o = stdout;
 	if(err)
 	{
@@ -1134,6 +1143,7 @@ static void long_usage(int err)
 	fprintf(o,"\ninput options\n\n");
 	fprintf(o," -k <n> --skip <n>         skip n frames at beginning\n");
 	fprintf(o," -n     --frames <n>       play only <n> frames of every stream\n");
+	fprintf(o,"        --fuzzy            Enable fuzzy seeks (guessing byte offsets or using approximate seek points from Xing TOC)\n");
 	fprintf(o," -y     --no-resync        DISABLES resync on error (--resync is deprecated)\n");
 	fprintf(o," -p <f> --proxy <f>        set WWW proxy\n");
 	fprintf(o," -u     --auth             set auth values for HTTP access\n");
@@ -1141,6 +1151,7 @@ static void long_usage(int err)
 	fprintf(o," -@ <f> --list <f>         play songs in playlist <f> (plain list, m3u, pls (shoutcast))\n");
 	fprintf(o," -l <n> --listentry <n>    play nth title in playlist; show whole playlist for n < 0\n");
 	fprintf(o,"        --loop <n>         loop track(s) <n> times, < 0 means infinite loop (not with --random!)\n");
+	fprintf(o,"        --keep-open        (--remote mode only) keep loaded file open after reaching end\n");
 #ifndef WIN32
 	fprintf(o,"        --timeout <n>      Timeout in seconds before declaring a stream dead (if <= 0, wait forever)\n");
 #endif
@@ -1148,6 +1159,8 @@ static void long_usage(int err)
 	fprintf(o," -Z     --random           full random play\n");
 	fprintf(o,"        --no-icy-meta      Do not accept ICY meta data\n");
 	fprintf(o," -i     --index            index / scan through the track before playback\n");
+	fprintf(o,"        --index-size <n>   change size of frame index\n");
+	fprintf(o,"        --preframes  <n>   number of frames to decode in advance after seeking (to keep layer 3 bit reservoir happy)\n");
 	fprintf(o,"        --resync-limit <n> Set number of bytes to search for valid MPEG data; <0 means search whole stream.\n");
 	fprintf(o,"\noutput/processing options\n\n");
 	fprintf(o," -o <o> --output <o>       select audio output module\n");
@@ -1169,8 +1182,8 @@ static void long_usage(int err)
 	fprintf(o,"        --force-3dnow      force use of 3DNow! optimized routine (obsoleted by --test-cpu)\n");
 	fprintf(o,"        --no-3dnow         force use of floating-pointer routine (obsoleted by --cpu)\n");
 	#endif
-	fprintf(o," -g     --gain             set audio hardware output gain\n");
-	fprintf(o," -f <n> --scale <n>        scale output samples (soft gain, default=%g)\n", (double)param.outscale);
+	fprintf(o," -g     --gain             [DEPRECATED] set audio hardware output gain\n");
+	fprintf(o," -f <n> --scale <n>        scale output samples (soft gain - based on 32768), default=%li)\n", param.outscale);
 	fprintf(o,"        --rva-mix,\n");
 	fprintf(o,"        --rva-radio        use RVA2/ReplayGain values for mix/radio mode\n");
 	fprintf(o,"        --rva-album,\n");
@@ -1184,6 +1197,9 @@ static void long_usage(int err)
 	fprintf(o," -4     --4to1             4:1 downsampling\n");
   fprintf(o,"        --pitch <value>    set hardware pitch (speedup/down, 0 is neutral; 0.05 is 5%%)\n");
 	fprintf(o,"        --8bit             force 8 bit output\n");
+	fprintf(o,"        --float            force floating point output (internal precision)\n");
+	audio_enclist(&enclist);
+	fprintf(o," -e <c> --encoding <c>     force a specific encoding c (%s)\n", enclist != NULL ? enclist : "OOM!");
 	fprintf(o," -d n   --doublespeed n    play only every nth frame\n");
 	fprintf(o," -h n   --halfspeed   n    play every frame n times\n");
 	fprintf(o,"        --equalizer        exp.: scales freq. bands acrd. to 'equalizer.dat'\n");
@@ -1196,6 +1212,7 @@ static void long_usage(int err)
 	fprintf(o," -o l   --lineout          (aix/hp/sun) output to lineout\n");
 #ifndef NOXFERMEM
 	fprintf(o," -b <n> --buffer <n>       set play buffer (\"output cache\")\n");
+	fprintf(o,"        --preload <value>  fraction of buffer to fill before playback\n");
 	fprintf(o,"        --smooth           keep buffer over track boundaries\n");
 #endif
 
@@ -1211,6 +1228,7 @@ static void long_usage(int err)
 	fprintf(o,"        --title            set xterm/rxvt title to filename\n");
 	#endif
 	fprintf(o,"        --long-tag         spacy id3 display with every item on a separate line\n");
+	fprintf(o,"        --utf8             Regardless of environment, print metadata in UTF-8.\n");
 	fprintf(o," -R     --remote           generic remote interface\n");
 	fprintf(o,"        --remote-err       force use of stderr for generic remote interface\n");
 #ifdef FIFO
