@@ -1,7 +1,7 @@
 /*
 	mpg123: main code of the program (not of the decoder...)
 
-	copyright 1995-2008 by the mpg123 project - free software under the terms of the LGPL 2.1
+	copyright 1995-2009 by the mpg123 project - free software under the terms of the LGPL 2.1
 	see COPYING and AUTHORS files in distribution or http://mpg123.org
 	initially written by Michael Hipp
 
@@ -13,71 +13,18 @@
 #define MPG123_H
 
 /* everyone needs it */
-#include "config.h"
-#include "httpget.h"
+#include "compat.h"
 /* import DLL symbols on windows */
 #if defined(WIN32) && defined(DYNAMIC_BUILD)
 #define LINK_MPG123_DLL
 #endif
 #include "mpg123.h"
-#include "compat.h"
 #define MPG123_REMOTE
 #define REMOTE_BUFFER_SIZE 2048
 #define MAXOUTBURST 32768
 
-#ifdef HAVE_STDLIB_H
-#include <stdlib.h>
-#endif
-
-#include        <stdio.h>
-#include        <string.h>
-#ifdef HAVE_SIGNAL_H
-#include        <signal.h>
-#endif
-#include        <math.h>
-
-#ifndef WIN32
-#include        <sys/signal.h>
-#endif
-/* For select(), I need select.h according to POSIX 2001, else: sys/time.h sys/types.h unistd.h */
-#ifdef HAVE_SYS_SELECT_H
-#include <sys/select.h>
-#endif
-#ifdef HAVE_SYS_TIME_H
-#include <sys/time.h>
-#endif
-#ifdef HAVE_UNISTD_H
-#include <unistd.h>
-#endif
-/* want to suport large files in future */
-#ifdef HAVE_SYS_TYPES_H
-#include <sys/types.h>
-#endif
-
-/* More integer stuff, just take what we can get... */
-#ifdef HAVE_INTTYPES_H
-#include <inttypes.h>
-#endif
-#ifdef HAVE_STDINT_H
-#include <stdint.h>
-#endif
-#ifdef HAVE_LIMITS_H
-#include <limits.h>
-#endif
- 
-#ifndef SIZE_MAX
-#define SIZE_MAX ((size_t)-1)
-#endif
-#ifndef ULONG_MAX
-#define ULONG_MAX ((unsigned long)-1)
-#endif
-
-#ifdef OS2
-#include <float.h>
-#endif
-
-typedef unsigned char byte;
 #include "xfermem.h"
+#include "httpget.h"
 
 #ifdef __GNUC__
 #define INLINE inline
@@ -85,14 +32,13 @@ typedef unsigned char byte;
 #define INLINE
 #endif
 
-#include "module.h"
 #include "audio.h"
+#include "local.h"
 
 extern size_t bufferblock;
 
 #define VERBOSE_MAX 3
 
-extern int utf8env; /* if we should print out UTF-8 or ASCII */
 extern char* binpath; /* argv[0], actually... */
 
 struct parameter
@@ -138,11 +84,7 @@ struct parameter
 	long doublespeed;
 	long start_frame;  /* frame offset to begin with */
 	long frame_number; /* number of frames to decode */
-#ifdef FLOATOUT
-	double outscale;
-#else
 	long outscale;
-#endif
 	int flags;
 	long force_rate;
 	int talk_icy;
@@ -151,6 +93,13 @@ struct parameter
 	double pitch; /* <0 or >0, 0.05 for 5% speedup. */
 	int ignore_mime; /* An mpg123 app flag field in future? */
 	char *proxyurl;
+	int keep_open; /* Whether to keep files open after end reached, for remote control mode, perhaps terminal control, too. */
+	int force_utf8; /* Regardless of environment, always print out verbatim UTF for metadata. */
+	long index_size; /* size of frame index */
+	char *force_encoding;
+	double preload; /* buffer preload size (fraction of full buffer) */
+	long preframes;
+	long gain; /* audio output gain, for selected outputs */
 };
 
 extern char *equalfile;
@@ -195,4 +144,8 @@ void prev_track(void);
 int  open_track(char *fname);
 void close_track(void);
 void set_intflag(void);
+
+/* equalizer... success is 0, failure -1 */
+int load_equalizer(mpg123_handle *mh);
+
 #endif 
