@@ -6,12 +6,11 @@
 	initially written by Michael Hipp
 */
 
-#include "config.h"
-
-
 #ifndef _MPG123_TERM_H_
 #define _MPG123_TERM_H_
 
+#include "mpg123app.h"
+
 #ifdef HAVE_TERMIOS
 
 #define LOOP_CYCLES	0.500000	/* Loop time in sec */
@@ -52,16 +51,26 @@
 #define PREV_KEY 'd'
 #define MPEG_KEY 'm'
 #define TAG_KEY  't'
-/* this counts as "undocumented" and can disappear */
+#define PITCH_UP_KEY    'c'
+#define PITCH_BUP_KEY   'C'
+#define PITCH_DOWN_KEY  'x'
+#define PITCH_BDOWN_KEY 'X'
+#define PITCH_ZERO_KEY  'w'
+/* This counts as "undocumented" and can disappear */
 #define FRAME_INDEX_KEY 'i'
+#define VARIOUS_INFO_KEY 'I'
+/* The normal and big pitch adjustment done on key presses. */
+#define PITCH_VAL 0.001
+#define PITCH_BVAL 0.01
 
 #define PAUSED_STRING	"Paused. \b\b\b\b\b\b\b\b"
 #define STOPPED_STRING	"Stopped.\b\b\b\b\b\b\b\b"
 #define EMPTY_STRING	"        \b\b\b\b\b\b\b\b"
 
 void term_init(void);
-off_t term_control(mpg123_handle *mh);
+off_t term_control(mpg123_handle *mh, audio_output_t *ao);
 void term_restore(void);
+void term_hint(void); /* Print a message hinting at terminal usage. */
 
 #endif
 
